package it.unibz.inf.completeness;

import java.util.List;
import java.util.Set;

import com.hp.hpl.jena.sparql.core.BasicPattern;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.query.Query;

import it.unibz.inf.semweb.QueryToolkit;

// completeness statements with no conditions aka Compl(P)
public class CompletenessStatement {

	private BasicPattern bgp;

	// constructors
	public CompletenessStatement(BasicPattern triples) {
		bgp = triples;
	}
	
	public CompletenessStatement(List<Triple> triples) {
		bgp = BasicPattern.wrap(triples);
	}

	// methods
	public BasicPattern getBGP() {
		return bgp;
	}

	public Query getSelectQuery() {
		return QueryToolkit.bgpToSelectQuery(bgp);
	}

	public Query getConstructQuery() {
		return QueryToolkit.bgpToConstructQuery(bgp);
	}
	
	public Set<String> getSignature() {
		return SignatureFactory.getSignature(bgp);
	}

	// for performance reason, toString() uses the BGP of the statement instead of other fancy representations
	public String toString() {
		return getBGP().toString();
	}
	
	// to calculate the hash key
	public int hashCode(){
	    return this.toString().hashCode();
	}
	
	// equals() is used to determine if a collection contains a given element
	// credit: http://tutorials.jenkov.com/java-collections/hashcode-equals.html
	public boolean equals(Object o){
	    if(o == null) return false;
	    
	    if(!(o instanceof CompletenessStatement)) return false;

	    CompletenessStatement other = (CompletenessStatement) o;
	    if(this.toString().compareTo(other.toString()) != 0) return false;

	    return true;
	  }

}
