package it.unibz.inf.completeness;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.sparql.core.BasicPattern;

import it.unibz.inf.semweb.*;

// constant map is a hashmap used in optimizing completeness reasoning without graphs (i.e., \C \models Compl(P))
public class ConstantMap {

	private HashMap<List<String>, List<CompletenessStatement>> hashmap = new HashMap<List<String>, List<CompletenessStatement>>();
	
	public ConstantMap(List<CompletenessStatement> csList) {
		hashmap = generateHashmap(csList);
	}
	
	public static List<String> getUniqueConstantList(BasicPattern bgp) {
		Set<String> constantSet = new HashSet<String>();
		for(Triple t : bgp) {
			Node s = t.getSubject();
			Node p = t.getPredicate();
			Node o = t.getObject();
			if(!s.isVariable())
				constantSet.add(s.toString());
			if(!p.isVariable())
				constantSet.add(p.toString());
			if(!o.isVariable())
				constantSet.add(o.toString());
		}
		List<String> constantList = new ArrayList<String>(constantSet);
		Collections.sort(constantList);
		return constantList;
	}
	
	public static List<String> getUniqueConstantList(Query q) {
		return getUniqueConstantList(QueryToolkit.body(q));
	}
	
	public static List<String> getUniqueConstantList(CompletenessStatement cs) {
		return getUniqueConstantList(cs.getBGP());
	}
	
	public static HashMap<List<String>, List<CompletenessStatement>> generateHashmap(List<CompletenessStatement> csList) {
		
		HashMap<List<String>, List<CompletenessStatement>> generatedHashmap = new HashMap<List<String>, List<CompletenessStatement>>();
		
		for(CompletenessStatement cs : csList) {
			
			// getUniqueConstantList() already returns a distinct and sorted list
			List<String> constantList = getUniqueConstantList(cs);
			if(generatedHashmap.containsKey(constantList))
				generatedHashmap.get(constantList).add(cs);
			else {
				List<CompletenessStatement> newCSList = new ArrayList<CompletenessStatement>();
				newCSList.add(cs);
				generatedHashmap.put(constantList, newCSList);
			}
			
		}
		
		return generatedHashmap;
		
	}
	
	public HashMap<List<String>, List<CompletenessStatement>> getHashmap() {
		return hashmap;
	}
	
	private List<CompletenessStatement> relevantCSList;
	
	public List<CompletenessStatement> search(List<String> constantListOfQuery) {
		relevantCSList = new ArrayList<CompletenessStatement>(); // reset the content of relevantCSList
		recSearch(new ArrayList<String>(), constantListOfQuery, 0);
		return relevantCSList;
	}

	// credit to Werner Nutt: idea is flipping bits between 0 and 1
	private void recSearch(List<String> prefix,
			List<String> predicateListOfQuery, int level) {
		
		if(level > predicateListOfQuery.size() - 1) {
			if(hashmap.containsKey(prefix))
				relevantCSList.addAll(hashmap.get(prefix));
		}
		else {
			List<String> nonGrowingPrefix = new ArrayList<String>();
			List<String> growingPrefix = new ArrayList<String>();
			nonGrowingPrefix.addAll(prefix);
			growingPrefix.addAll(prefix);
			growingPrefix.add(predicateListOfQuery.get(level));
			recSearch(nonGrowingPrefix, predicateListOfQuery, level + 1);
			recSearch(growingPrefix, predicateListOfQuery, level + 1);
		}
		
	}
		
}
