package it.unibz.inf.soundness.experiment;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

// class for an experiment observation
public class Observation {

	// query evaluation
	String queryString;
	int queryLength;
	int queryResultsSize;
	List<Long> queryEvaluations;
	
	// answer soundness reasoning
	int soundAnswersSize;
	List<Long> queryEvaluationsInAnswerSoundnessChecks;
	List<Long> answerSoundnessChecksWithoutQueryEvaluation;
	List<Long> answerSoundnessChecksWithoutQueryEvaluationNaive;
	List<Long> answerSoundnessChecks; // this will be the query evaluation time plus answer soundness reasoning time
	
	// pattern soundness reasoning
	boolean isSound;
	List<Long> patternSoundnessChecks;
	List<Long> patternSoundnessChecksNaive;
	
	// constructor
	public Observation(String queryString, int queryLength) {

		this.queryString = queryString;
		this.queryLength = queryLength;
		this.queryResultsSize = 0;
		this.queryEvaluations = new ArrayList<Long>();
		
		this.soundAnswersSize = 0;
		this.queryEvaluationsInAnswerSoundnessChecks = new ArrayList<Long>();
		this.answerSoundnessChecksWithoutQueryEvaluation = new ArrayList<Long>();
		this.answerSoundnessChecksWithoutQueryEvaluationNaive = new ArrayList<Long>();
		this.answerSoundnessChecks = new ArrayList<Long>();
		
		this.isSound = false;
		this.patternSoundnessChecks = new ArrayList<Long>();
		this.patternSoundnessChecksNaive = new ArrayList<Long>();

	}

	//////////////////////////////
	// query evaluation methods
    //////////////////////////////
	
	public void setQueryResultsSize(int queryResultsSize) {
		this.queryResultsSize = queryResultsSize;
	}

	public List<Long> getQueryEvaluations() {
		return queryEvaluations;
	}
	
	public long getMedianQueryEvaluations() {
		return medianOf(queryEvaluations);
	}
	
	//////////////////////////////
	// answer soundness reasoning methods
    //////////////////////////////
	
	public void setSoundAnswersSize(int soundAnswersSize) {
		this.soundAnswersSize = soundAnswersSize;
	}
	
	public List<Long> getQueryEvaluationsInAnswerSoundnessChecks() {
		return queryEvaluationsInAnswerSoundnessChecks;
	}

	public long getMedianQueryEvaluationsInAnswerSoundnessChecks() {
		return medianOf(queryEvaluationsInAnswerSoundnessChecks);
	}
	
	public List<Long> getAnswerSoundnessChecksWithoutQueryEvaluation() {
		return answerSoundnessChecksWithoutQueryEvaluation;
	}
	
	public long getMedianAnswerSoundnessChecksWithoutQueryEvaluation() {
		return medianOf(answerSoundnessChecksWithoutQueryEvaluation);
	}
	
	public List<Long> getAnswerSoundnessChecksWithoutQueryEvaluationNaive() {
		return answerSoundnessChecksWithoutQueryEvaluationNaive;
	}
	
	public long getMedianAnswerSoundnessChecksWithoutQueryEvaluationNaive() {
		return medianOf(answerSoundnessChecksWithoutQueryEvaluationNaive);
	}
	
	public List<Long> getAnswerSoundnessChecks() {
		return answerSoundnessChecks;
	}
	
	public long getMedianAnswerSoundnessChecks() {
		return medianOf(answerSoundnessChecks);
	}
	
	//////////////////////////////
	// pattern soundness reasoning methods
    //////////////////////////////
	
	public void setPatternSoundnessFlag(boolean isSound) {
		this.isSound = isSound;		
	}
	
	public List<Long> getPatternSoundnessChecks() {
		return patternSoundnessChecks;
	}

	public long getMedianPatternSoundnessChecks() {
		return medianOf(patternSoundnessChecks);
	}
	
	public List<Long> getPatternSoundnessChecksNaive() {
		return patternSoundnessChecksNaive;
	}

	public long getMedianPatternSoundnessChecksNaive() {
		return medianOf(patternSoundnessChecksNaive);
	}
	
	//////////////////////////////
	// standard methods
    //////////////////////////////
	
	public String toString() {
		String outputString = "";
		outputString = outputString + queryLength;
		outputString = outputString + ", " + queryResultsSize;
		outputString = outputString + ", " + getMedianQueryEvaluations();
		outputString = outputString + ", " + soundAnswersSize;
		outputString = outputString + ", " + getMedianQueryEvaluationsInAnswerSoundnessChecks();
		outputString = outputString + ", " + getMedianAnswerSoundnessChecksWithoutQueryEvaluation();
		//outputString = outputString + ", " + getMedianAnswerSoundnessChecksWithoutQueryEvaluationNaive();
		outputString = outputString + ", " + getMedianAnswerSoundnessChecks();
		outputString = outputString + ", " + isSound;
		outputString = outputString + ", " + getMedianPatternSoundnessChecks();
		//outputString = outputString + ", " + getMedianPatternSoundnessChecksNaive();
		return outputString;
	}

	//////////////////////////////
	// static methods
    //////////////////////////////
	
	// given observations, print them all
	public static void printAll(List<Observation> observations) {
		for(Observation observation : observations)
			System.out.println(observation.toString());
	}
	
	public static Long medianOf(List<Long> resultList) {
		
		Collections.sort(resultList);
		int listSize = resultList.size();
		Double middleIndex = Math.floor(listSize/2);
		
		if(listSize == 0) return Long.valueOf(0);
		else {
			if(listSize % 2 == 0) // if even
				return ( (resultList.get(middleIndex.intValue() - 1) + resultList.get(middleIndex.intValue())) 
						  /2 );
			else // if odd
				return resultList.get(middleIndex.intValue());
		}
		
	}
	
}
