package it.unibz.inf.soundness.experiment;

import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import com.hp.hpl.jena.sparql.engine.binding.Binding;

import it.unibz.inf.semweb.*;
import it.unibz.inf.soundness.*;
import it.unibz.inf.completeness.*;

// main class for running soundness experiments
public class SoundnessExperimentWithNaivePatternChecking {

	// for reliability, we take X repeats for each experiment observation
	static int observationRepeats = 10;

	static boolean warmingUpFlag = true;
	
	static int experimentCase = 1;
	
	// default TDB directory
	static String tdbDir = "tdb-PREPROCESSED-wikidata-simple-statements-20160201";

	// queries file
	static String queriesFile = "test-queries.txt";

	// parameters: [observation repeats] [warming up] [case] [query filename]
	// [observation repeats]: how many should an observation is repeated, e.g., 1, 10, 20
	// [warming up]: should a warming up be performed? i.e., 0 (no), 1 (yes)
	// [case]: which case should be performed?
	//     1 : oneTP
	//     2 : oneTPoneTP
	//     3 : twoTPsTO
	//     4 : twoTPsSE
	// [query filename]: what is the query filename e.g., "test-queries.txt"
	public static void main(String[] args) {
		
		// credit: http://stackoverflow.com/questions/15571496/how-to-check-if-a-folder-exists
		if (!Files.isDirectory(Paths.get(tdbDir))) {
			tdbDir = "C:\\Users\\user\\Documents\\Datasets\\tdb-PREPROCESSED-wikidata-simple-statements-20160201";
		}
		
		if(args.length > 0) {
			observationRepeats = Integer.parseInt(args[0]);
			if(Integer.parseInt(args[1]) == 0)
				warmingUpFlag = false;
			experimentCase = Integer.parseInt(args[2]);
			queriesFile = args[3];
		}

		System.out.println("##### OBSERVATION REPEAT(S): " + observationRepeats);
		System.out.println("##### WITH WARMING UP: " + warmingUpFlag);
		System.out.println("##### EXPERIMENT CASE: " + experimentCase);
		System.out.println("##### QUERIES FILENAME: " + queriesFile);
		
		long startTime;
		long endTime;
		long elapsedTime;
		
		List<QueryWithNegation> queriesWithNegation;
		if(experimentCase == 1) {
			queriesWithNegation = QueryToolkit.genQueriesWithNegationOneTP(QueryToolkit.readQueries(queriesFile));
		}
		else if(experimentCase == 2) {
			queriesWithNegation = QueryToolkit.genQueriesWithNegationOneTPOneTP(QueryToolkit.readQueries(queriesFile));
		}
		
		else if(experimentCase == 5) {
			queriesWithNegation = QueryToolkit.genQueriesWithNegationThreeTPs(QueryToolkit.readQueries(queriesFile));
		}
		
		else {	
			queriesWithNegation = QueryToolkit.genQueriesWithNegationTwoTPs(QueryToolkit.readQueries(queriesFile));
		}
		
		System.out.println("##### QUERIES:\n" + queriesWithNegation + "\n");

		List<Observation> observations = new ArrayList<Observation>();
		for(QueryWithNegation queryWithNegation : queriesWithNegation) {
			observations.add(new Observation(queryWithNegation.toString(), queryWithNegation.getLength()));
		}
		
		// DEBUG
//		QueryWithNegation queryNeg = queriesWithNegation.get(0);
//		Model prototypicalGraph = QueryToolkit.freeze(queryNeg.posPart);
//		prototypicalGraph.add(QueryToolkit.freeze(queryNeg.negParts.get(0)));
//		System.out.println(prototypicalGraph);
//		System.out.println(QueryToolkit.defreeze(prototypicalGraph));
//		System.exit(0);
		
		CompletenessReasoner reasoner = new CompletenessReasoner(tdbDir);
		
		//////////////////////////////////////////////////
		// QUERY EVALUATION EXPERIMENTS
		//////////////////////////////////////////////////
		
		if(warmingUpFlag) {
		
			// WARMING UP 3 TIMES
			System.out.println("##### WARMING UP FOR QUERY EVALUATIONS:");
			for(QueryWithNegation queryWithNegation : queriesWithNegation) {
				System.out.println("Evaluating:\n" + queryWithNegation);
				reasoner.evalSELECT(queryWithNegation.getQuery());
				reasoner.evalSELECT(queryWithNegation.getQuery());
				reasoner.evalSELECT(queryWithNegation.getQuery());
			}
		
		}

		System.out.println("##### QUERY EVALUATIONS:");
		
		reasoner.close();
		for(int i = 0; i < observationRepeats; i++) {
			for(int j = 0; j < queriesWithNegation.size(); j++) {
				QueryWithNegation queryWithNegation = queriesWithNegation.get(j);
				
				// DEBUG
				System.out.println(queryWithNegation);
				
				reasoner = new CompletenessReasoner(tdbDir);
				startTime = System.nanoTime();
				List<Binding> queryEvalResults = reasoner.evalSELECT(queryWithNegation.getQuery());
				endTime = System.nanoTime();
				elapsedTime = endTime-startTime;
				int queryResultsSize = queryEvalResults.size();
				System.out.println(queryResultsSize + " results, with elapsed time: " + elapsedTime);
				observations.get(j).getQueryEvaluations().add(elapsedTime);
				observations.get(j).setQueryResultsSize(queryResultsSize);
				reasoner.close();
			}
			System.out.println("---");
		}
		reasoner = new CompletenessReasoner(tdbDir);
		System.out.println();
				
		//////////////////////////////////////////////////
		// ANSWER SOUNDNESS EXPERIMENTS
		//////////////////////////////////////////////////
		
		System.out.println("##### GENERATING COMPLETENESS TEMPLATES:");
		
		List<CompletenessTemplate> answerSoundnessTemplates;
		if(experimentCase != 4) {
			answerSoundnessTemplates = reasoner.genCompletenessTemplatesForAnswerSoundness(queriesWithNegation);
		} else {
			answerSoundnessTemplates = reasoner.genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(queriesWithNegation);
		}
		
		System.out.println("\n##### GENERATED COMPLETENESS TEMPLATES: " + answerSoundnessTemplates.size() + "\n");

		answerSoundnessTemplates = CompletenessTemplate.removeRedundancies(answerSoundnessTemplates);
		System.out.println("##### NON-REDUNDANT TEMPLATES: " + answerSoundnessTemplates.size() + "\n");
		
		// DEBUG
//		for(CompletenessTemplate template : answerSoundnessTemplates) {
//			List<CompletenessStatement> statementsOfTemplate = template.unfold();
//			System.out.println(template + " : " + statementsOfTemplate.size() + " : " + template.bindings.size() + " : " + template.bindings);
//		}
//		reasoner.close();
//		System.exit(0);
		
		HashMap<String, List<CompletenessTemplate>> templateMap = SignatureFactory.getSignatureMapForTemplates(answerSoundnessTemplates);
		
		// free up memory
		answerSoundnessTemplates = new ArrayList<CompletenessTemplate>();
		
		if(warmingUpFlag) {
		
			// WARMING UP 3 TIMES
			System.out.println("##### WARMING UP ANSWER SOUNDNESS:\n");
			reasoner.close();
			for(QueryWithNegation queryWithNegation : queriesWithNegation) {
					reasoner = new CompletenessReasoner(tdbDir);
					System.out.println("Checking answer soundness of:\n" + queryWithNegation);
					SoundnessReasoner.getSoundAnswersOptimized(queryWithNegation.posPart, queryWithNegation.negParts, templateMap, reasoner);
//					SoundnessReasoner.getSoundAnswersOptimized(queryWithNegation.posPart, queryWithNegation.negParts, templateMap, reasoner);
//					SoundnessReasoner.getSoundAnswersOptimized(queryWithNegation.posPart, queryWithNegation.negParts, templateMap, reasoner);
					reasoner.close();
			}
			reasoner = new CompletenessReasoner(tdbDir);

		}
		
		System.out.println("##### ANSWER SOUNDNESS CHECKING:");

		reasoner.close();
		for(int i = 0; i < observationRepeats; i++) {
			for(int j = 0; j < queriesWithNegation.size(); j++) {
				QueryWithNegation queryWithNegation = queriesWithNegation.get(j);
				reasoner = new CompletenessReasoner(tdbDir);
				
				long queryEvalStartTime = System.nanoTime();
				reasoner.evalSELECT(queryWithNegation.getQuery());
				long queryEvalEndTime = System.nanoTime();
				long queryEvalElapsedTime = queryEvalEndTime - queryEvalStartTime;
				
				long answerSoundnessCheckStartTime = System.nanoTime();
//				List<Binding> soundAnswers = SoundnessReasoner.getSoundAnswersOptimizedFromBindings(answersToCheck, queryWithNegation.negParts, templateMap, reasoner);
				List<Binding> soundAnswers = new ArrayList<Binding>();
				long answerSoundnessCheckEndTime = System.nanoTime();
				long answerSoundnessCheckElapsedTime = answerSoundnessCheckEndTime - answerSoundnessCheckStartTime;
				
				long queryEvalAndAnswerSoundnessTime = queryEvalElapsedTime + answerSoundnessCheckElapsedTime;
				
				int soundAnswersSize = soundAnswers.size();
		
				System.out.println("#Sound answers: " + soundAnswersSize + ", with Query Evaluation time: " + queryEvalElapsedTime + ", with Answer Soundness Check time: " + answerSoundnessCheckElapsedTime + ", with total time: " + queryEvalAndAnswerSoundnessTime);
				observations.get(j).getQueryEvaluationsInAnswerSoundnessChecks().add(queryEvalElapsedTime);
				observations.get(j).getAnswerSoundnessChecksWithoutQueryEvaluation().add(answerSoundnessCheckElapsedTime);
				observations.get(j).getAnswerSoundnessChecks().add(queryEvalAndAnswerSoundnessTime);
				observations.get(j).setSoundAnswersSize(soundAnswersSize);
				reasoner.close();
			}
			System.out.println("---");
		}
		reasoner = new CompletenessReasoner(tdbDir);
		System.out.println();
		
		// DEBUG
//		Observation.printAll(observations);
//		reasoner.close();
//		System.exit(0);
		
		//////////////////////////////////////////////////
		// PATTERN SOUNDNESS EXPERIMENTS
		//////////////////////////////////////////////////
		
		// free memory
		templateMap = new HashMap<String, List<CompletenessTemplate>>();
		
		System.out.println("##### GENERATING COMPLETENESS STATEMENTS:");
		
		List<CompletenessTemplate> patternSoundnessTemplates = new ArrayList<CompletenessTemplate>();
				
		if(experimentCase != 4) {
			answerSoundnessTemplates = reasoner.genCompletenessTemplatesForAnswerSoundness(queriesWithNegation);
		} else {
			answerSoundnessTemplates = reasoner.genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(queriesWithNegation);
		}
			
		answerSoundnessTemplates = CompletenessTemplate.removeRedundancies(answerSoundnessTemplates);		

		patternSoundnessTemplates.addAll(answerSoundnessTemplates);
		patternSoundnessTemplates.addAll(reasoner.genCompletenessTemplatesForPatternSoundness(queriesWithNegation));

		// DEBUG
//		for(int i = 0; i < 1000; i++)
//			patternSoundnessTemplates.addAll(reasoner.genCompletenessTemplatesForPatternSoundness(queriesWithNegation));

		List<CompletenessStatement> statements = new ArrayList<CompletenessStatement>();
		for(CompletenessTemplate template : patternSoundnessTemplates) {
			List<CompletenessStatement> statementsOfTemplate = template.unfold();
			statements.addAll(statementsOfTemplate);
			
			// DEBUG
//			System.out.println(statementsOfTemplate.size() + " : " + template.bindings.size());
			
		}
		
		// free memory
		patternSoundnessTemplates = new ArrayList<CompletenessTemplate>();
		
		System.out.println("\n##### GENERATED COMPLETENESS STATEMENTS FOR PATTERN SOUNDNESS: " + statements.size() + "\n");
		ConstantMap constantMap = new ConstantMap(statements);
		
		// naive pattern soundness checking
//		statements = new ArrayList<CompletenessStatement>(); // free up some memory

		if(warmingUpFlag) {
		
			// WARMING UP 3 TIMES
			System.out.println("##### WARMING UP PATTERN SOUNDNESS:\n");
			reasoner.close();
			for(QueryWithNegation queryWithNegation : queriesWithNegation) {
				reasoner = new CompletenessReasoner(tdbDir);
				System.out.println("Checking pattern soundness of:\n" + queryWithNegation);
				SoundnessReasoner.checkSchemaSoundnessOptimized(queryWithNegation.posPart, queryWithNegation.negParts, constantMap, reasoner);
				SoundnessReasoner.checkSchemaSoundnessOptimized(queryWithNegation.posPart, queryWithNegation.negParts, constantMap, reasoner);
				SoundnessReasoner.checkSchemaSoundnessOptimized(queryWithNegation.posPart, queryWithNegation.negParts, constantMap, reasoner);
				reasoner.close();
			}
			reasoner = new CompletenessReasoner(tdbDir);

		}
		
		System.out.println("##### PATTERN SOUNDNESS CHECKING:");
		reasoner.close();
		for(int i = 0; i < observationRepeats; i++) {
			for(int j = 0; j < queriesWithNegation.size(); j++) {
				QueryWithNegation queryWithNegation = queriesWithNegation.get(j);
				reasoner = new CompletenessReasoner(tdbDir);
				startTime = System.nanoTime();
				boolean isSound = SoundnessReasoner.checkSchemaSoundnessOptimized(queryWithNegation.posPart, queryWithNegation.negParts, constantMap, reasoner);
				endTime = System.nanoTime();
				elapsedTime = endTime-startTime;
				System.out.println("#Sound graph pattern: " + isSound + ", with time elapsed " + elapsedTime);
				observations.get(j).setPatternSoundnessFlag(isSound);
				observations.get(j).getPatternSoundnessChecks().add(elapsedTime);
				
				// naive pattern soundness checking
				startTime = System.nanoTime();
				isSound = SoundnessReasoner.checkSchemaSoundnessNaive(queryWithNegation.posPart, queryWithNegation.negParts, statements, reasoner);
				endTime = System.nanoTime();
				elapsedTime = endTime-startTime;
				System.out.println("# (NAIVE) Sound graph pattern: " + isSound + ", with time elapsed " + elapsedTime);
				observations.get(j).getPatternSoundnessChecksNaive().add(elapsedTime);
				
				reasoner.close();
			}
			System.out.println("---");
		}
		reasoner = new CompletenessReasoner(tdbDir);
		System.out.println();
		
		System.out.println("##### STATISTICS:");
		Observation.printAll(observations);
		
		reasoner.close();
	}
	
}
